"""Run the testing routing based on the `scenery` settings.

See: `python -m scenery --help`
"""



if __name__ == "__main__":

    import sys

    import scenery.cli


    exit_code = scenery.cli.main()
    sys.exit(exit_code)



import argparse
import importlib
import os
from pathlib import Path
import sys
import typing


from scenery.common import summarize_test_result, interpret, iter_on_manifests
import scenery.cli
from scenery import logger


########################
# SCENERY CONFIG
########################

def scenery_setup(args: argparse.Namespace) -> bool:
    """Read the settings module and set the corresponding environment variables.

    This function imports the specified settings module and sets environment variables
    based on its contents. The following environment variables are set:

    SCENERY_COMMON_ITEMS
    SCENERY_SET_UP_INSTRUCTIONS
    SCENERY_TESTED_APP_NAME
    SCENERY_MANIFESTS_FOLDER

    Args:
        settings_location (str): The location (import path) of the settings module.

    Raises:
        ImportError: If the settings module cannot be imported.
    """

    sys.path.append(os.path.join('.'))
    settings = importlib.import_module(args.scenery_settings_module)
    
    # Env variables
    os.environ["SCENERY_COMMON_ITEMS"] = settings.SCENERY_COMMON_ITEMS
    os.environ["SCENERY_SET_UP_INSTRUCTIONS"] = settings.SCENERY_SET_UP_INSTRUCTIONS
    os.environ["SCENERY_POST_REQUESTS_INSTRUCTIONS_SELENIUM"] = (
        settings.SCENERY_POST_REQUESTS_INSTRUCTIONS_SELENIUM
    )
    os.environ["SCENERY_TESTED_APP_NAME"] = settings.SCENERY_TESTED_APP_NAME
    os.environ["SCENERY_MANIFESTS_FOLDER"] = settings.SCENERY_MANIFESTS_FOLDER

    emojy, msg, color, log_lvl = interpret(True)
    logger.info("scenery set-up", style=color)
    
    return True

###################
# DJANGO CONFIG
###################

def django_setup(args: argparse.Namespace) -> bool:
    """Set up the Django environment.

    This function sets the DJANGO_SETTINGS_MODULE environment variable and calls django.setup().

    Args:
        settings_module (str): The import path to the Django settings module.
    """
    import django

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", args.django_settings_module)
    logger.debug(f"{os.environ.get('DJANGO_SETTINGS_MODULE')=}")
    django.setup()

    
    from django.conf import settings as django_settings
    logger.debug(f"{django_settings.INSTALLED_APPS=}")

    emojy, msg, color, log_lvl = interpret(True)
    logger.log(log_lvl, "django set-up", style=color)
    
    return True


###################
# INTEGRATION TESTS
###################

def integration_tests(args: argparse.Namespace) -> bool:
    """
    Execute the main functionality of the scenery test runner.

    Returns:
        exit_code (int): Exit code indicating success (0) or failure (1)
    """
    # NOTE mad: this needs to be loaded afeter scenery_setup and django_setup
    from scenery.core import process_manifest_as_integration_test


    # FIXME mad: this is here to be able to load driver in two places
    # See also core.TestsLoader.tests_from_manifest.
    # Probably not a great pattern but let's fix this later
    # driver = get_selenium_driver(headless=args.headless)
    driver = None

    report_data  : dict[str, typing.List[typing.Tuple[bool, dict]]] = {
        "dev_backend": [],
        "dev_frontend": [],
        "remote_backend": [],
        "remote_frontend": []
    }


    for filename in iter_on_manifests(args):
        
        results = process_manifest_as_integration_test(filename, args=args, driver=driver)

        for key, val in results.items():
            if val:
                success, summary = summarize_test_result(val, key.replace("_", "-"))
                report_data[key].append((success, summary))

    success = scenery.cli.report_integration(report_data)

    return success



###################
# LOAD TESTS
###################

def load_tests(args: argparse.Namespace) -> bool:
    # NOTE mad: this needs to be loaded after scenery_setup and django_setup
    from scenery.core import process_manifest_as_load_test

    success = True
    report_data : dict[str, typing.List] = {}

    for filename in iter_on_manifests(args):        
        results = process_manifest_as_load_test(filename, args=args)
        file_level_success = scenery.cli.report_load(results)
        report_data.update(results)
        success &= file_level_success

    return success


###################
# CODE
###################

def inspect_code(args: argparse.Namespace) -> bool:
    from scenery.inspect_code import count_line_types

    report_data = {}

    # Get all files recursively
    folder = Path(args.folder)

    # TODO: just the directory (make option)

    # All files
    for file_path in folder.rglob('*.py'):
        if file_path.is_file():
            report_data[str(file_path)] = count_line_types(file_path)

    success = scenery.cli.report_inspect(report_data)

    return success




"""A versatile integration testing framework for django apps."""

import logging
from typing import Any

from rich.console import Console

class SceneryLogger:

    style_map = {
        logging.DEBUG: None,
        logging.INFO: "blue",
        logging.WARNING: "yellow",
        logging.ERROR: "red",
    }

    def __init__(self, level: int):
        self.level = level
        self.console = console
        # pass

    def log(self, level: int, msg: Any, style: str | None = None) -> None:
        if self.level <= level:
            level_name = f"{logging.getLevelName(level)}"
            level_name = f"{level_name:<10}"
            color = self.style_map[level]
            if color:
                level_name = f"[{color}]{level_name}[/{color}]"
                msg = level_name + str(msg)
            self.console.log(msg, style=style)

    def info(self, msg: Any, style: str | None =None) -> None:
        self.log(logging.INFO, msg, style)

    def debug(self, msg: Any, style: str | None =None) -> None:
        self.log(logging.DEBUG, msg, style)

    def warning(self, msg: Any, style: str | None =None) -> None:
        self.log(logging.WARNING, msg, style)

    def error(self, msg: Any, style: str | None =None) -> None:
        self.log(logging.ERROR, msg, style)
        
console = Console()
logger = SceneryLogger(logging.INFO)




# NOTE mad: do not erase

# def show_root_logger_config():
#     """Show detailed configuration of the root logger only."""
#     root = logging.getLogger()
    
#     print("🌳 ROOT LOGGER CONFIGURATION")
#     print("=" * 50)
#     print(f"Name: '{root.name}' (empty = root)")
#     print(f"Level: {logging.getLevelName(root.level)} ({root.level})")
#     print(f"Effective Level: {logging.getLevelName(root.getEffectiveLevel())}")
#     print(f"Propagate: {root.propagate}")
#     print(f"Disabled: {root.disabled}")
#     print(f"Number of Handlers: {len(root.handlers)}")
    
#     if root.handlers:
#         print(f"\n📋 HANDLERS:")
#         for i, handler in enumerate(root.handlers):
#             print(f"  [{i}] {type(handler).__name__}")
#             print(f"      Level: {logging.getLevelName(handler.level)} ({handler.level})")
            
#             # Formatter info
#             if hasattr(handler, 'formatter') and handler.formatter:
#                 fmt = handler.formatter
#                 print(f"      Format: '{fmt._fmt}'")
#                 print(f"      Date Format: '{fmt.datefmt}'")
#             else:
#                 print(f"      Formatter: None")
            
#             # Stream info for StreamHandler
#             if hasattr(handler, 'stream'):
#                 stream_name = getattr(handler.stream, 'name', str(handler.stream))
#                 print(f"      Stream: {stream_name}")
            
#             # File info for FileHandler
#             elif hasattr(handler, 'baseFilename'):
#                 print(f"      File: {handler.baseFilename}")
            
#             # Rich-specific info
#             if 'Rich' in type(handler).__name__:
#                 print(f"      Rich Tracebacks: {getattr(handler, 'rich_tracebacks', 'N/A')}")
#                 print(f"      Markup: {getattr(handler, 'markup', 'N/A')}")
#                 if hasattr(handler, 'console') and handler.console:
#                     print(f"      Console File: {getattr(handler.console, 'file', 'N/A')}")
#     else:
#         print("\n📋 No handlers configured")



"""Execute instructions used in `TestCase.setUp` and `TestCase.setUpTestData` provided in the manifest."""

import importlib
import os
from typing import Callable

from scenery import logger
from scenery.common import SceneryTestCase, RemoteBackendTestCase, LoadTestCase
from scenery.manifest import SetUpInstruction


def local_execution_only(func: Callable) -> Callable:
    """Simple decorator that marks function as executable"""
    # NOTE mad: this type ignore is here to stay, this is the whole
    # point to create a non existing attribute.
    func._local_execution_only = True # type: ignore [attr-defined]
    return func

class SetUpHandler:
    """Responsible for executing instructions used in `TestCase.setUp` and `TestCase.setUpTestData` provided in the manifest.

    This class dynamically imports and executes setup instructions specified in the test manifest.
    It is typically used by the MethodBuilder to construct setup methods for test cases.

    Attributes:
        module: The imported module containing setup instruction implementations.
        logger: A logger instance for debug output.
    """
    
    instructions_module = importlib.import_module(os.environ["SCENERY_SET_UP_INSTRUCTIONS"])
    # selenium_module = importlib.import_module(os.environ["SCENERY_SET_UP_INSTRUCTIONS_SELENIUM"])

    @staticmethod
    def exec_set_up_instruction(
        # NOTE mad: it either takes the instance or the class
        # depending whether it is class method or not
        # (setUp vs. setUpClass)
        testcase: SceneryTestCase | type[SceneryTestCase],
        instruction: SetUpInstruction,
    ) -> None:
        """Execute the method corresponding to the SetUpInstruction.

        This method dynamically retrieves and executes the setup function specified
        by the SetUpInstruction. It logs the execution for debugging purposes.

        Args:
            testcase (SceneryTestCase): The Scenery test case class or instance.
            instruction (scenery.manifest.SetUpInstruction): The setup instruction to execute.

        Raises:
            AttributeError: If the specified setup function is not found in the imported module.
        """
        func = getattr(SetUpHandler.instructions_module, instruction.command)


        if isinstance(testcase, (RemoteBackendTestCase, LoadTestCase)) and hasattr(func, "_local_execution_only"):
            pass
        else:
            logger.debug(instruction)
            func(testcase, **instruction.args)




"""Parse manifest YAML files and build proper Manifests."""

import os
import io
from typing import Any, cast

import scenery.common
import scenery.manifest

import yaml
from yaml.constructor import ConstructorError


#####################
# PARSER
#####################


class ManifestParser:
    """
    A class responsible for parsing test manifest files in YAML format.

    This class provides methods to validate, format, and parse manifest files
    into Python objects that can be used by the testing framework.

    Attributes:
        common_items (dict): Common items loaded from a YAML file specified by the SCENERY_COMMON_ITEMS environment variable.
    """

    common_items: dict[str, dict] = scenery.common.read_yaml(os.environ["SCENERY_COMMON_ITEMS"])

    ##########
    # RAW DICT
    ##########

    @staticmethod
    def validate_dict(d: dict) -> None:  # scenery.manifest.RawManifestDict
        """
        Validate the top-level keys of a manifest dictionary.

        This method checks if only valid keys are present at the top level and ensures
        that either singular or plural forms of 'case' and 'scene' are provided, but not both.

        Args:
            d (dict): The manifest dictionary to validate.

        Raises:
            ValueError: If invalid keys are present or if the case/scene keys are not correctly specified.
        """
        if not all(
            key in scenery.manifest.RawManifestDict.__annotations__.keys() for key in d.keys()
        ):
            raise ValueError(
                f"Invalid key(s) in {d.keys()} ({d.get('manifest_origin', 'No origin found.')})"
            )

        for key in ["case", "scene"]:
            has_one = key in d
            has_many = f"{key}s" in d

            if has_one and has_many:
                raise ValueError(
                    f"Both `{key}` and `{key}s` keys are present at top level.",
                )

            if key == "scene" and not (has_one or has_many):
                raise ValueError(
                    f"Neither `{key}` and `{key}s` keys are present at top level.",
                )

    @staticmethod
    def format_dict(d: scenery.manifest.RawManifestDict) -> scenery.manifest.ManifestDict:
        """
        Reformat the manifest dictionary to ensure it has all expected keys and provide default values if needed.

        Args:
            d (dict): The original manifest dictionary.

        Returns:
            dict: A formatted dictionary with all expected keys.
        """
        return {
            "set_up_class": d.get("set_up_class", []),
            "set_up": d.get("set_up", []),
            "scenes": ManifestParser._format_dict_scenes(d),
            "cases": ManifestParser._format_dict_cases(d),
            "manifest_origin": d["manifest_origin"],
            "ttype": d.get("ttype")
        }

    @staticmethod
    def _format_dict_cases(d: scenery.manifest.RawManifestDict) -> dict[str, dict]:
        has_one = "case" in d
        has_many = "cases" in d
        if has_one:
            return {"CASE": d["case"]}
        elif has_many:
            return d["cases"]
        else:
            return {"NO_CASE": {}}

    @staticmethod
    def _format_dict_scenes(d: scenery.manifest.RawManifestDict) -> list[dict]:
        has_one = "scene" in d
        has_many = "scenes" in d
        if has_one:
            return [d["scene"]]
        elif has_many:
            return d["scenes"]
        else:
            raise ValueError

    @staticmethod
    def parse_dict(d: dict) -> scenery.manifest.Manifest:
        """
        Parse a manifest dictionary into a Manifest object.

        This method validates the dictionary, formats it, and then parses it into a Manifest object.

        Args:
            d (dict): The manifest dictionary to parse.

        Returns:
            scenery.manifest.Manifest: A Manifest object created from the input dictionary.
        """
        ManifestParser.validate_dict(d)
        formatted_d = ManifestParser.format_dict(cast(scenery.manifest.RawManifestDict, d))
        return scenery.manifest.Manifest.from_formatted_dict(formatted_d)

    ##########
    # YAML
    ##########

    @staticmethod
    def validate_yaml(obj: Any) -> None:
        """
        Validate the structure of a YAML-loaded manifest.

        This method checks if the YAML content is a dictionary and if it contains only expected keys.

        Args:
            obj (Any): The YAML content to validate.

        Raises:
            TypeError: If the YAML content is not a dictionary.
            ValueError: If the YAML content contains unexpected keys.
        """
        if not isinstance(obj, dict):
            raise TypeError(f"Manifest need to be a dict not a '{type(obj)}'")

        if not all(
            key in scenery.manifest.RawManifestDict.__annotations__.keys() for key in obj.keys()
        ):
            raise ValueError(
                f"Invalid key(s) in {obj.keys()} ({obj.get('manifest_origin', 'No origin found.')})"
            )

    @staticmethod
    def _yaml_constructor_case(
        loader: yaml.SafeLoader, node: yaml.nodes.Node
    ) -> scenery.manifest.Substituable:
        if isinstance(node, yaml.nodes.ScalarNode):
            return scenery.manifest.Substituable(loader.construct_scalar(node))
        else:
            raise ConstructorError

    @staticmethod
    def _yaml_constructor_common_item(loader: yaml.SafeLoader, node: yaml.nodes.Node) -> dict:
        if isinstance(node, yaml.nodes.ScalarNode):
            return ManifestParser.common_items[loader.construct_scalar(node)]
        if isinstance(node, yaml.nodes.MappingNode):
            d = loader.construct_mapping(node)
            case = ManifestParser.common_items[d["ID"]] | {
                key: value for key, value in d.items() if key != "ID"
            }
            return case
        else:
            raise ConstructorError

    @staticmethod
    # def read_manifest_yaml(filename: str) -> Any:
    def read_manifest_yaml(stream: str | io.TextIOWrapper) -> Any:
        """
        Read a YAML manifest stream with custom tags.

        This method uses a custom YAML loader to handle special tags like !case and !common-item.

        Args:
            stream(str | StringIO): The stream of the YAML manifest to read.

        Returns:
            dict: The parsed content of the YAML file.
        """
        # NOTE: inspired by https://matthewpburruss.com/post/yaml/

        # Add constructor
        Loader = yaml.FullLoader
        Loader.add_constructor("!case", ManifestParser._yaml_constructor_case)
        Loader.add_constructor("!common-item", ManifestParser._yaml_constructor_common_item)

        # with open(filename) as f:
        #     content = yaml.load(f, Loader)

        content = yaml.load(stream, Loader)
        ManifestParser.validate_yaml(content)
        return content


    @staticmethod
    def parse_yaml_from_file(filename: str) -> scenery.manifest.Manifest:
        """
        Parse a YAML manifest file into a Manifest object.

        This method reads the YAML file, validates its content, and then parses it into a Manifest object.

        Args:
            filename (str): The filename of the YAML manifest to parse.

        Returns:
            scenery.manifest.Manifest: A Manifest object created from the YAML file.
        """
        with open(filename) as f:
            d = ManifestParser.read_manifest_yaml(f)
        d["manifest_origin"] = d.get("manifest_origin", filename)
        return ManifestParser.parse_dict(d)




"""General functions and classes used by other modules."""

import argparse
from collections import Counter
import os
import logging
import typing
import unittest
from typing import TypeVar, Union
import requests

import django
from django.test.runner import DiscoverRunner as DjangoDiscoverRunner
from django.contrib.staticfiles.testing import StaticLiveServerTestCase

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
# from selenium.webdriver.chrome.service import Service

from scenery import console, logger

import yaml


###################
# SELENIUM
###################


def get_selenium_driver(headless: bool) -> webdriver.Chrome:
    """Return a Selenium WebDriver instance configured for Chrome."""
    chrome_options = Options()
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--disable-dev-shm-usage')
    chrome_options.add_argument('--disable-gpu')
    chrome_options.add_argument('--window-size=1920,1080')
    chrome_options.add_argument('--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36')
    if headless:
        chrome_options.add_argument("--headless=new")         # NOTE mad: For newer Chrome versions
        # chrome_options.add_argument("--headless")           
    driver = webdriver.Chrome(options=chrome_options)  #  service=service
    driver.implicitly_wait(10)
    return driver


###################
# CLASSES
###################

# NOTE mad: this is here to prevent circular import and still use those types
# in reponse_checker and toher scripts for instance. 

class DjangoBackendTestCase(django.test.TestCase):
    """A Django TestCase for backend testing."""


class DjangoFrontendTestCase(StaticLiveServerTestCase):
    """A Django TestCase for frontend testing."""
    base_url: str
    driver: webdriver.Chrome

class RemoteBackendTestCase(unittest.TestCase):
    """A TestCase for backend testing on a remote server."""
    mode: str
    session: requests.Session
    base_url: str
    headers: dict[str, str]

class RemoteFrontendTestCase(unittest.TestCase):
    """A TestCase for backend testing on a remote server."""
    mode: str
    driver: webdriver.Chrome
    # session: requests.Session
    base_url: str
    headers: dict[str, str]



class LoadTestCase(unittest.TestCase):
    """A TestCase for load testing on a remote server."""
    mode: str
    session: requests.Session
    headers: dict[str, str]
    base_url: str
    data: dict[str, typing.List[dict[str, int|float]]]
    users:int
    requests_per_user:int


SceneryTestCaseTypes = Union[DjangoBackendTestCase, DjangoFrontendTestCase, RemoteBackendTestCase, RemoteFrontendTestCase, LoadTestCase]
SceneryTestCase = TypeVar("SceneryTestCase", bound=SceneryTestCaseTypes)


# NOTE mad: could be useful when we fit FastAPI
# DjangoTestCaseTypes = Union[DjangoBackendTestCase, DjangoFrontendTestCase]
# DjangoTestCase = TypeVar("DjangoTestCase", bound=DjangoTestCaseTypes)

########
# YAML #
########


def read_yaml(filename: str) -> typing.Any:
    """Read and parse a YAML file.

    Args:
        filename (str): The path to the YAML file to be read.

    Returns:
        Any: The parsed content of the YAML file.

    Raises:
        yaml.YAMLError: If there's an error parsing the YAML file.
        IOError: If there's an error reading the file.
    """
    with open(filename, "r") as f:
        return yaml.safe_load(f)


def iter_on_manifests(args: argparse.Namespace) -> typing.Iterable[str]:
    for filename in os.listdir(os.environ["SCENERY_MANIFESTS_FOLDER"]):
        if args.manifest is not None and filename.replace(".yml", "") != args.manifest:
            continue

        yield filename



##################
# UNITTEST
##################


def serialize_unittest_result(result: unittest.TestResult) -> Counter:
    """Serialize a unittest.TestResult object into a dictionary.

    Args:
        result (unittest.TestResult): The TestResult object to serialize.

    Returns:
        dict: A dictionary containing the serialized TestResult data.
    """
    d = {
        attr: getattr(result, attr)
        for attr in [
            "failures",
            "errors",
            "testsRun",
            "skipped",
            "expectedFailures",
            "unexpectedSuccesses",
        ]
    }
    d = {key: len(val) if isinstance(val, list) else val for key, val in d.items()}
    return Counter(d)


def summarize_test_result(result: unittest.TestResult, test_label: str) -> tuple[bool, Counter]:
    """Return true if the tests all succeeded, false otherwise."""
    for failed_test, traceback in result.failures:
        test_name = failed_test.id()
        emojy, msg, color, log_lvl = interpret(False)
        # logger.log(log_lvl, f"[{color}]{test_name} {msg}[/{color}]\n{traceback}")
        logger.log(log_lvl, f"{test_name} {msg}", style=color)
        # console.print_exception(
        #     exc_info=(None, None, traceback),
        #     show_locals=True
        # )
        console.print(traceback)

    for failed_test, traceback in result.errors:
        test_name = failed_test.id()
        emojy, msg, color, log_lvl = interpret(False)
        # logger.log(log_lvl, f"[{color}]{test_name} {msg}[/{color}]\n{traceback}")
        logger.log(log_lvl, f"{test_name} {msg}", style=color)
        console.print(traceback)

    success = True
    summary = serialize_unittest_result(result)
    if summary["errors"] > 0 or summary["failures"] > 0:
        success = False
    if summary["testsRun"] == 0:
        pass

    else:
        emojy, msg, color, log_lvl = interpret(success)
        msg = f"{test_label} {msg}"
        logger.log(log_lvl, msg, style=color)

    return success, summary


def interpret(success: bool) -> typing.Tuple[str, str, str, int]:
    """Return emojy, success/failure message, color and log level corresponding to the succes of an execution."""
    if success:
        emojy, msg, color, log_lvl = "🟢", "passed", "green", logging.INFO
    else:
        emojy, msg, color, log_lvl = "❌", "failed", "red", logging.ERROR
    return emojy, msg, color, log_lvl


###################
# DJANGO TEST
###################


def overwrite_get_runner_kwargs(
    django_runner: DjangoDiscoverRunner, stream: typing.IO
) -> dict[str, typing.Any]:
    """Overwrite the get_runner_kwargs method of Django's DiscoverRunner.

    This function is used to avoid printing Django test output by redirecting the stream.

    Args:
        django_runner (DiscoverRunner): The Django test runner instance.
        stream: The stream to redirect output to.

    Returns:
        dict: A dictionary of keyword arguments for the test runner.

    Notes:
        see django.test.runner.DiscoverRunner.get_runner_kwargs
    """
    kwargs = {
        "failfast": django_runner.failfast,
        "resultclass": django_runner.get_resultclass(),
        "verbosity": django_runner.verbosity,
        "buffer": django_runner.buffer,
        # NOTE: this is the line below that changes compared to the original
        "stream": stream,
    }
    return kwargs


class CustomDiscoverRunner(DjangoDiscoverRunner):
    """Custom test runner that allows for stream capture."""
    
    # NOTE mad: this was done to potentially shut down the original stream
    # NOTE mad: used both in rehearsal and core module (for the runner
    # TODO mad: once we fit FastAPI, this runner should only be used 
    # for the django backend and frontend test

    def __init__(self, stream: typing.Any , *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)
        self.stream = stream

    # def __del__(self):
    #     print(self.stream.getvalue())

    def get_test_runner_kwargs(self) -> dict[str, typing.Any]:
        """Overwrite the original from django.test.runner.DiscoverRunner."""
        return overwrite_get_runner_kwargs(self, self.stream)



"""Represent all data conveied by the manifest."""

from dataclasses import dataclass, field
import enum
import os
import http
from typing import Any
from urllib.parse import urlparse
import re
import typing
import itertools

from django.utils.http import urlencode
from django.urls.exceptions import NoReverseMatch
from django.db.models.base import ModelBase

from django.apps import apps as django_apps
from django.urls import reverse 



#############################
# TYPE HINTS
#############################


class RawManifestDict(typing.TypedDict, total=False):
    """The type of dict which can appear in the yamls."""

    set_up_class: typing.Sequence[dict]
    set_up: typing.Sequence[dict]
    case: dict
    cases: dict[str, dict]
    scene: dict
    scenes: typing.List[dict]
    manifest_origin: str
    ttype: str


class ManifestDict(typing.TypedDict):
    """The clean version of the dict parsed, ready to be transformed into a proper manifest."""

    scenes: typing.List[dict]
    cases: dict[str, dict]
    manifest_origin: str
    set_up_class: typing.Sequence[str | dict]
    set_up: typing.Sequence[str | dict]
    ttype: typing.Optional[str]


########################
# SINGLE KEY DICTIONNARY
########################


SingleKeyDictKey = typing.TypeVar("SingleKeyDictKey", bound=str)
SingleKeyDictKeyValue = typing.TypeVar("SingleKeyDictKeyValue")


@dataclass
class SingleKeyDict(typing.Generic[SingleKeyDictKey, SingleKeyDictKeyValue]):
    """A dataclass representing a dictionary with a single key-value pair.

    This class is useful for having a quick as_tuple representation of a dict {key:value}
    returned as (key, value).

    Attributes:
        _dict (Dict[SingleKeyDictKey, SingleKeyDictKeyValue]): The underlying dictionary.
        key (SingleKeyDictKey): The single key in the dictionary.
        value (SingleKeyDictKeyValue): The value associated with the single key.

    Methods:
        validate(): Ensures the dictionary has exactly one key-value pair.
        as_tuple(): Returns the key-value pair as a tuple.

    Raises:
        ValueError: If the dictionary does not contain exactly one key-value pair.
    """

    _dict: typing.Dict[SingleKeyDictKey, SingleKeyDictKeyValue] = field()
    key: SingleKeyDictKey = field(init=False)
    value: SingleKeyDictKeyValue = field(init=False)

    def __post_init__(self) -> None:
        self.validate()
        self.key, self.value = next(iter(self._dict.items()))

    def validate(self) -> None:
        """Check the dictonary has indeed a single key."""
        if len(self._dict) != 1:
            raise ValueError(
                f"SingleKeyDict should have length 1 not '{len(self._dict)}'\n{self._dict}"
            )

    def as_tuple(self) -> tuple:
        """🔴 This should not be confused with built-in method datclasses.astuple."""
        return self.key, self.value


####################################
# ENUMS
####################################


########
# CHECKS
########


class DirectiveCommand(enum.Enum):
    """Values allowed for Manifest["checks"]."""

    STATUS_CODE = "status_code"
    REDIRECT_URL = "redirect_url"
    COUNT_INSTANCES = "count_instances"
    DOM_ELEMENT = "dom_element"
    FIELD_OF_INSTANCE = "field_of_instance"
    # JS_VARIABLE = "js_variable"
    JS_STRINGIFY = "js_stringify"
    JSON = "inspect_json"


class DomArgument(enum.Enum):
    """Values allowed for Manifest["checks]["dom_element"]."""

    FIND = "find"
    FIND_ALL = "find_all"
    COUNT = "count"
    SCOPE = "scope"
    TEXT = "text"
    ATTRIBUTE = "attribute"


##########################
# SET UP TEST DATA, SET UP
##########################


@dataclass(frozen=True)
class SetUpInstruction:
    """Store the command and potential arguments for setUpTestData and setUp."""

    command: str
    args: dict[str, Any] = field(default_factory=dict)

    @classmethod
    def from_object(cls, x: str | dict) -> "SetUpInstruction":
        """Return an instruction from a string or a dict."""
        match x:
            case str(s):
                cmd_name, args = s, {}
            case dict(d) if len(d) == 1:
                cmd_name, args = SingleKeyDict(d).as_tuple()
            case dict(d):
                raise ValueError(
                    f"`SetUpInstruction` cannot be instantiated from dictionnary of length {len(x)}\n{x}"
                )
            case _:
                raise TypeError(f"`SetUpInstruction` cannot be instantiated from {type(x)}")

        return cls(cmd_name, args)


########################
# CASE
########################


@dataclass(frozen=True)
class Item:
    """Store potential information that will be used to build the HTTP Request."""

    _id: str
    _dict: dict[str, Any]

    def __getitem__(self, key: str) -> Any:
        return self._dict[key]


@dataclass(frozen=True)
class Case:
    """Store a collection of items representing a test case.

    Attributes:
        _id (str): The identifier for this case.
        items (dict[str, Item]): A dictionary of items in this case, indexed by item ID.

    Methods:
        __getitem__(item_id): Retrieve an item from the case by its ID.

    Class Methods:
        from_id_and_dict(case_id: str, items: dict[str, dict]) -> Case:
            Create a Case instance from a case ID and a dictionary of items.
    """

    _id: str
    items: dict[str, Item]

    def __getitem__(self, item_id: str) -> Item:
        return self.items[item_id]

    @classmethod
    def from_id_and_dict(cls, case_id: str, items_dict: dict[str, dict[str, Any]]) -> "Case":
        """Return a case from an id and a dict."""
        items = {item_id: Item(item_id, item_dict) for item_id, item_dict in items_dict.items()}
        return cls(case_id, items)


########################
# SCENES
########################


@dataclass
class Substituable:
    """Represent the field which need to be replace by some value coming from a given case."""

    field_repr: str
    regex_field = re.compile(r"^(?P<item_id>[a-z_]+):?(?P<field_name>[a-z_]+)?$")

    def __post_init__(self) -> None:
        if not (re_match := re.match(self.regex_field, self.field_repr)):
            raise ValueError(f"Invalid field representation '{self.field_repr}'")
        else:
            self.field_repr_match = re_match

    def shoot(self, case: Case) -> Any:
        """Return the corresponding case's value based on the field representation."""
        match self.field_repr_match.groups():
            case item_id, None:
                # There is only a reference to the item
                # In this case , we pass all the variables with the dict
                # return case[item_id][self.target]
                return case[item_id]._dict
            case item_id, field_name:
                # There is only a reference to the item:field_name
                # In this case, we pass only the corresponding value
                # field_value = case[item_id][self.target][field_name]
                field_value = case[item_id][field_name]
                return field_value


@dataclass
class Directive:
    """Store a given check to perform, before the substitution (this is part of a Scene, not a Take).

    This class represents a directive (check) to be performed on an HTTP response,
    before case-specific substitutions have been made.

    Attributes:
        instruction (DirectiveCommand): The type of check to perform.
        args (Any): The arguments for the check.

    Class Methods:
        from_dict(directive_dict: dict) -> Directive:
            Create an Directive instance from a dictionary.
    """

    instruction: DirectiveCommand
    args: Any

    def __post_init__(self) -> None:
        """Format self.args."""
        match self.instruction, self.args:
            case DirectiveCommand.STATUS_CODE, int(n):
                self.args = http.HTTPStatus(n)
                # NOTE: Workaround if we want the class to be frozen
                # object.__setattr__(self, "args", HTTPStatus(n))
                pass
            case DirectiveCommand.STATUS_CODE, Substituable():
                pass
            case DirectiveCommand.DOM_ELEMENT, dict(d):
                self.args = {DomArgument(key): value for key, value in d.items()}
                # Check if there is and only one locator
                locators = [
                    self.args.get(key, None) for key in (DomArgument.FIND_ALL, DomArgument.FIND)
                ]
                if not any(locators):
                    raise ValueError("Neither `find_all` or `find` provided to check DOM element")
                else:
                    locators = list(filter(None, locators))
                    if len(locators) > 1:
                        raise ValueError("More than one locator provided")

            case DirectiveCommand.DOM_ELEMENT, Substituable():
                pass
            case DirectiveCommand.REDIRECT_URL, str(s):
                pass
            case DirectiveCommand.REDIRECT_URL, Substituable():
                pass
            case DirectiveCommand.COUNT_INSTANCES, {"model": str(s), "n": int(n)}:
                # if os.environ.get("DJANGO_SETTINGS_MODULE"):
                app_config = django_apps.get_app_config(os.environ["SCENERY_TESTED_APP_NAME"])
                self.args["model"] = app_config.get_model(s)
            case DirectiveCommand.COUNT_INSTANCES, Substituable():
                pass
            case DirectiveCommand.JSON, {"key": str(s), "value": _}: # 
                pass
            case DirectiveCommand.FIELD_OF_INSTANCE, {"find": {"model": str(model)}, "field": str(s), "value": _}: # 
                # if os.environ.get("DJANGO_SETTINGS_MODULE"):
                app_config = django_apps.get_app_config(os.environ["SCENERY_TESTED_APP_NAME"])
                self.args["find"]["model"] = app_config.get_model(model)
            case _:
                raise ValueError(
                    f"Cannot interpret '{self.instruction}:({self.args})' as Directive"
                )

    @classmethod
    def from_dict(cls, directive_dict: dict) -> "Directive":
        """Return the Directive based on the provided dict."""
        instruction, args = SingleKeyDict(directive_dict).as_tuple()
        return cls(DirectiveCommand(instruction), args)


@dataclass
class Scene:
    """Store all actions to perform, before the substitution of information from the `Cases`.

    This class represents an Scene, which includes the method, URL, and various
    parameters and checks to be performed.

    Attributes:
        method (http.HTTPMethod): The HTTP method for this scene.
        url (str): The URL or URL pattern for this scene.
        directives (list[Directive]): The list of directives (checks) to perform.
        data (dict[str, Any]): The data to be sent with the request.
        query_parameters (dict): Query parameters for the URL.
        url_parameters (dict): URL parameters for reverse URL lookup.

    Methods:
        shoot(case: Case) -> Take:
            Create an Take instance by substituting case values into the scene.

    Class Methods:
        from_dict(d: dict) -> Scene:
            Create an Scene instance from a dictionary.
        substitute_recursively(x, case: Case):
            Recursively substitute values from a case into a data structure.
    """

    method: http.HTTPMethod
    url: str
    directives: list[Directive]
    data: dict[str, Any] = field(default_factory=dict)
    query_parameters: dict = field(default_factory=dict)
    url_parameters: dict = field(default_factory=dict)
    actions: typing.Optional[list[SetUpInstruction]] = field(default_factory=list)


    def __post_init__(self) -> None:
        self.method = http.HTTPMethod(self.method)
        # At this point we don't check url as we wait for subsitution
        # potentially occuring through data/query_parameters/url_parameters

    @classmethod
    def from_dict(cls, d: dict) -> "Scene":
        """Return a scene from a dict."""
        d["directives"] = [Directive.from_dict(directive) for directive in d["directives"]]
        return cls(**d)

    @classmethod
    def substitute_recursively(cls, x: typing.Any, case: Case) -> typing.Any:
        """Perform the substitution."""
        match x:
            case int(_) | str(_):
                return x
            case ModelBase():
                return x
            case Substituable(_):
                return x.shoot(case)
            case Directive(instruction, args):
                return Check(instruction, cls.substitute_recursively(args, case))
            case dict(_):
                return {key: cls.substitute_recursively(value, case) for key, value in x.items()}
            case list(_):
                return [cls.substitute_recursively(value, case) for value in x]
            case _:
                raise NotImplementedError(f"Cannot substitute recursively '{x}' ('{type(x)}')")

    def shoot(self, case: Case) -> "Take":
        """Return the Take resulting from the case applied to its scene."""
        return Take(
            method=self.method,
            url=self.url,
            query_parameters=self.query_parameters,
            data=self.substitute_recursively(self.data, case),
            url_parameters=self.substitute_recursively(self.url_parameters, case),
            checks=self.substitute_recursively(self.directives, case),
        )


################
# MANIFEST
################


@dataclass(frozen=True)
class Manifest:
    """Store all the information to build/shoot all different `Takes`.

    This class represents a complete test manifest, including setup instructions,
    test cases, and scenes to be executed.

    Attributes:
        set_up_test_data (list[SetUpInstruction]): Instructions for setting up test data.
        set_up (list[SetUpInstruction]): Instructions for general test setup.
        scenes (list[Scene]): The scenes to be executed.
        cases (dict[str, Case]): The test cases, indexed by case ID.
        manifest_origin (str): The origin of the manifest file.

    Class Methods:
        from_formatted_dict(d: dict) -> Manifest:
            Create a Manifest instance from a formatted dictionary.
    """

    set_up_class: list[SetUpInstruction]
    set_up: list[SetUpInstruction]
    scenes: list[Scene]
    cases: dict[str, Case]
    manifest_origin: str
    ttype: str | None

    @classmethod
    def from_formatted_dict(cls, d: ManifestDict) -> "Manifest":
        """Return a manifest from a dict with expected keys."""
        return cls(
            [
                SetUpInstruction.from_object(instruction)
                for instruction in d[
                    "set_up_class"
                ]  # d[ManifestFormattedDictKeys.set_up_test_data]
            ],
            [
                SetUpInstruction.from_object(instruction)
                for instruction in d["set_up"]  # d[ManifestFormattedDictKeys.set_up]
            ],
            [
                Scene.from_dict(scene) for scene in d["scenes"]
            ],  # d[ManifestFormattedDictKeys.scenes]],
            {
                case_id: Case.from_id_and_dict(case_id, case_dict)
                for case_id, case_dict in d[
                    "cases"
                ].items()  # d[ManifestFormattedDictKeys.cases].items()
            },
            # d[ManifestFormattedDictKeys.manifest_origin],
            d["manifest_origin"],
            d.get("ttype")
        )
    
    def iter_on_takes(
        self, 
        only_url: str | None, 
        only_case_id: str | None, 
        only_scene_pos: str | None
    ) -> typing.Iterable[typing.Tuple[str, int, "Take"]]:
        for (case_id, case), (scene_pos, scene) in itertools.product(
            self.cases.items(), enumerate(self.scenes)
        ):
            if only_case_id is not None and case_id != only_case_id:
                continue
            elif only_scene_pos is not None and str(scene_pos) != only_scene_pos:
                continue
            if only_url is not None and only_url != scene.url:
                continue
            take = scene.shoot(case)
            yield case_id, scene_pos, take


########################
# TAKES
########################


@dataclass
class Check(Directive):
    """Store a given check to perform (after the subsitution)."""

    def __post_init__(self) -> None:
        """Format self.args."""
        match self.instruction, self.args:
            case DirectiveCommand.STATUS_CODE, int(n):
                self.args = http.HTTPStatus(n)
            case DirectiveCommand.DOM_ELEMENT, dict(d):
                self.args = {DomArgument(key): value for key, value in d.items()}
                if attribute := self.args.get(DomArgument.ATTRIBUTE):
                    if value:= attribute.get("value"):
                        self.args[DomArgument.ATTRIBUTE]["value"] = (
                            self._format_dom_element_attribute_value(value)
                        )
            case DirectiveCommand.REDIRECT_URL, str(_):
                pass
            case DirectiveCommand.COUNT_INSTANCES, {"model": ModelBase(), "n": int(n)}:
                # NOTE mad: Validate model is registered
                app_config = django_apps.get_app_config(os.environ["SCENERY_TESTED_APP_NAME"])
                app_config.get_model(self.args["model"].__name__)
            case DirectiveCommand.JS_STRINGIFY, _:
                # TODO mad: js_stringify
                pass
            case DirectiveCommand.JSON, {"key": str(_), "value": _} :
                pass
            case DirectiveCommand.FIELD_OF_INSTANCE, {"find":{"model": ModelBase()}, "field": str(_), "value": _}:
                # NOTE mad: Validate model is registered
                app_config = django_apps.get_app_config(os.environ["SCENERY_TESTED_APP_NAME"])
                app_config.get_model(self.args["find"]["model"].__name__)
            
            case _:
                raise ValueError(
                    f"Cannot interpret '{self.instruction}:({self.args})' as Check"
                )

    @staticmethod
    def _format_dom_element_attribute_value(value: str | int | list[str]) -> list[str] | str:
        if isinstance(value, (str, list)):
            return value
        elif isinstance(value, int):
            return str(value)
        else:
            raise TypeError(
                f"attribute value can only be `str` or `list[str]` not {value} ('{type(value)}')"
            )


@dataclass
class Take:
    """Store all the information after the substitution from the `Cases` has been performed.

    This class represents a fully resolved HTTP request to be executed, including
    the method, URL, data, and checks to be performed.

    Attributes:
        method (http.HTTPMethod): The HTTP method for this take.
        url (str): The fully resolved URL for this take.
        checks (list[Check]): The list of checks to perform on the response.
        data (dict): The data to be sent with the request.
        query_parameters (dict): Query parameters for the URL.
        url_parameters (dict): URL parameters used in URL resolution.

    Notes:
        The `url` is expected to be either a valid URL or a registered viewname.
        If it's a viewname, it will be resolved using Django's `reverse` function.
    """

    method: http.HTTPMethod
    url: str
    checks: list[Check]
    data: dict
    query_parameters: dict
    url_parameters: dict

    def __post_init__(self) -> None:
        self.method = http.HTTPMethod(self.method)

        try:
            # First we try if the url is a django viewname
            url_name = self.url
            self.url = reverse(self.url, kwargs=self.url_parameters)
            self.url_name = url_name
        except NoReverseMatch:
            # Otherwise we check it is a valid url
            parsed = urlparse(self.url)
            if not (parsed.scheme and parsed.netloc):
                raise ValueError(f"'{self.url}' could not be reversed and is not a valid url")

        if self.query_parameters:
            # NOTE mad: We use http.urlencode instead for compatibility
            # https://stackoverflow.com/questions/4995279/including-a-querystring-in-a-django-core-urlresolvers-reverse-call
            # https://gist.github.com/benbacardi/227f924ec1d9bedd242b
            self.url += "?" + urlencode(self.query_parameters)
